#include <syslog.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "led.h"
#include "log.h"
#include "server.h"
#include "gpio.h"
#include "gpio_shift_register.h"

gpio_shift_register_t led_gpio_shift_register;
static const char *sk9822_attr_file = NULL;
static const char *sk9822_brightness_file = NULL;
#define SK9822_DEFAULT_BRIGHTNESS 15 /* Needs to be in sync with the gryphon kernel module */

void gpio_led_init(struct server_ctx *s_ctx);

typedef enum {
	LOW,
	HI,
	UNKNOWN,
} active_t;

typedef enum {
	MODE_UNKNOWN,
	DIRECT,
	SHIFTREG_GPIO,
	GPIO_LINUX,
	LED_LINUX,
	LED_SK9822,
} gpio_mode_t;

struct gpio_led_data {
	int addr;
	int delay;		/* time to wait for a gpio set to complete, HW workaround, do not set normally */
	union {
		struct {
			char *sysfs_attr;
			int brightness;
		}; /* LED_LINUX attributes */
		char *rgb_color; /* LED_SK9822 attribute */
	};
	active_t active;
	int state;
	gpio_mode_t mode;
	struct led_drv led;
};

/* Linux LED */
static int led_linux_set(char *sysfs_attr, int val)
{
	char buf[128] = "[none]", filename[128] = {0};
	int fd = -1;

	snprintf(filename, sizeof(filename), "/sys/class/leds/%s/trigger", sysfs_attr);
	fd = open(filename, O_RDONLY);
	if (fd >= 0) {
		snprintf(buf, sizeof(buf), "%d\n", val);
		read(fd, buf, sizeof(buf));
		close(fd);
	}

	if (!strncmp(buf, "[none]", 6) || (strstr(buf, "[test-mode]") != NULL)) {
		snprintf(filename, sizeof(filename), "/sys/class/leds/%s/brightness", sysfs_attr);

		fd = open(filename, O_WRONLY);
		if (fd >= 0) {
			snprintf(buf, sizeof(buf), "%d\n", val);
			write(fd, buf, strlen(buf));
			close(fd);
		}
	}

	return 0;
}

/* SK9822 LED */
static int led_sk9822_set(const char *rgb_val)
{
	int fd;

	/* sanity check */
	if (!(sk9822_attr_file && rgb_val))
		return -1;

	fd = open(sk9822_attr_file, O_WRONLY);
	if (fd != -1) {
		write(fd, rgb_val, 6);
		close(fd);
	}

	return 0;
}

static int led_sk9822_set_brightness(unsigned int brightness)
{
	int fd;
	char hex_bright[2] = {0};

	/* sanity check */
	if (!(sk9822_brightness_file))
		return -1;

	fd = open(sk9822_brightness_file, O_WRONLY);
	if (fd != -1) {
		snprintf(hex_bright, sizeof(hex_bright), "%x", brightness);
		write(fd, hex_bright, strlen(hex_bright));
		close(fd);
	}

	return 0;
}

static int gpio_led_set_state(struct led_drv *drv, led_state_t state)
{
	struct gpio_led_data *p = (struct gpio_led_data *)drv->priv;
	int bit_val = 0;

	if(state) {
		if(p->active)
			bit_val=1;
	} else {
		if(!p->active)
			bit_val=1;
	}

	p->state = state;

	switch (p->mode) {
	case DIRECT:
		gpio_set_state(p->addr, bit_val,p->delay);
		break;
	case GPIO_LINUX:
		gpio_linux_set(p->addr, bit_val);
		break;
	case SHIFTREG_GPIO:
		gpio_shift_register_cached_set(&led_gpio_shift_register, p->addr, bit_val);
		break;
	case LED_LINUX:
		{
			int val;
			val = state?p->brightness:0;
			if (p->active == LOW)
				val = 255 - val;
			led_linux_set(p->sysfs_attr, val);
		}
		break;
	case LED_SK9822:
		led_sk9822_set(state?p->rgb_color:"000000");
		led_sk9822_set_brightness(state?p->brightness:0);
		break;
	default:
		DBG(1,"access mode not supported [%d,%s]", p->mode, p->led.name);
	}

	return p->state;
}

static led_state_t gpio_led_get_state(struct led_drv *drv)
{
	struct gpio_led_data *p = (struct gpio_led_data *)drv->priv;
	DBG(1, "state for %s",  drv->name);

	return p->state;
}

static int gpio_led_support(struct led_drv *drv, led_state_t state)
{
	switch (state) {

	case OFF:
	case ON:
		return 1;
		break;

	default:
		return 0;
	}
	return 0;
}

static 	int led_set_brightness(struct led_drv *drv, int level)
{
	struct gpio_led_data *data = (struct gpio_led_data*)drv->priv;

	if (level > 100 || level < 0) {
		DBG(1, "invalid brightness level %d", level);
		return level;
	}

	if (data->mode == LED_LINUX) {
		data->brightness = (255 * level) / 100;
		DBG(2, "set brightness to %d for %s", data->brightness, data->led.name);
	} else if (data->mode == LED_SK9822) {
		data->brightness = (SK9822_DEFAULT_BRIGHTNESS * level) / 100;
		DBG(2, "set brightness to %d for %s", data->brightness, data->led.name);
	}

	return level;
}

static	int led_get_brightness(struct led_drv *drv)
{
	struct gpio_led_data *data = (struct gpio_led_data*)drv->priv;
	if (data->mode == LED_SK9822) {
		return (data->brightness * 100) / SK9822_DEFAULT_BRIGHTNESS;
	} else {
		return (data->brightness * 100) / 255;
	}
}

/* Linux LED */
static int led_linux_exit_testmode(char *sysfs_attr)
{
	char buf[128] = {0};
	int fd = -1;

	snprintf(buf, sizeof(buf), "/sys/class/leds/%s/active", sysfs_attr);

	fd = open(buf, O_WRONLY);
	if (fd >= 0) {
		strncpy(buf, "0", sizeof(buf));
		write(fd, buf, strlen(buf));
		close(fd);
		fd = -1;
	}

	snprintf(buf, sizeof(buf), "/sys/class/leds/%s/trigger", sysfs_attr);

	fd = open(buf, O_WRONLY);
	if (fd >= 0) {
		strncpy(buf, "none", sizeof(buf));
		write(fd, buf, strlen(buf));
		close(fd);
		return 0;
	}


	return 1;
}

static int led_exit_testmode(struct led_drv *drv)
{
	struct gpio_led_data *p = (struct gpio_led_data *)drv->priv;
	int rc = 1;

	switch (p->mode) {
	case LED_LINUX:
		{
			rc = led_linux_exit_testmode(p->sysfs_attr);
		}
		break;
	default:
		DBG(1,"test mode not supported [%d,%s]", p->mode, p->led.name);
	}

	return rc;
}



/* Linux LED */
static int led_linux_enter_testmode(char *sysfs_attr)
{
	char buf[128] = {0};
	int fd = -1;

	led_linux_exit_testmode(sysfs_attr);

	snprintf(buf, sizeof(buf), "/sys/class/leds/%s/trigger", sysfs_attr);

	fd = open(buf, O_WRONLY);
	if (fd >= 0) {
		strncpy(buf, "test-mode", sizeof(buf));
		write(fd, buf, strlen(buf));
		close(fd);
		return 0;
	}

	return 1;
}

static int led_enter_testmode(struct led_drv *drv)
{
	struct gpio_led_data *p = (struct gpio_led_data *)drv->priv;
	int rc = 1;

	switch (p->mode) {
	case LED_LINUX:
		{
			rc = led_linux_enter_testmode(p->sysfs_attr);
		}
		break;
	default:
		DBG(1,"test mode not supported [%d,%s]", p->mode, p->led.name);
	}

	return rc;
}

static struct led_drv_func func = {
	.set_state = gpio_led_set_state,
	.get_state = gpio_led_get_state,
	.set_brightness = led_set_brightness,
	.get_brightness = led_get_brightness,
	.enter_testmode = led_enter_testmode,
	.exit_testmode = led_exit_testmode,
	.support   = gpio_led_support,
};

void gpio_led_init(struct server_ctx *s_ctx) {

	LIST_HEAD(leds);
	struct ucilist *node;
	int gpio_shiftreg_clk=-1, gpio_shiftreg_dat=-1, gpio_shiftreg_mask=-1, gpio_shiftreg_bits=0;
	const char *s;

	DBG(1, "");

	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "gpio_shiftreg_clk")))
		gpio_shiftreg_clk = strtol(s,0,0);
	DBG(1, "gpio_shiftreg_clk = [%d]", gpio_shiftreg_clk);
	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "gpio_shiftreg_dat")))
		gpio_shiftreg_dat = strtol(s,0,0);
	DBG(1, "gpio_shiftreg_dat = [%d]", gpio_shiftreg_dat);
	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "gpio_shiftreg_mask")))
		gpio_shiftreg_mask = strtol(s,0,0);
	DBG(1, "gpio_shiftreg_mask = [%d]", gpio_shiftreg_mask);
	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "gpio_shiftreg_bits")))
		gpio_shiftreg_bits = strtol(s,0,0);
	DBG(1, "gpio_shiftreg_bits = [%d]", gpio_shiftreg_bits);
	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "sk9822_attr_file")))
		sk9822_attr_file = strdup(s);
	DBG(1, "sk9822_attr_file = [%s]", sk9822_attr_file);
	if((s=ucix_get_option(s_ctx->uci_ctx, "hw", "board", "sk9822_brightness_file")))
		sk9822_brightness_file = strdup(s);
	DBG(1, "sk9822_brightness_file = [%s]", sk9822_brightness_file);

	ucix_get_option_list(s_ctx->uci_ctx, "hw" ,"gpio_leds", "leds", &leds);
	list_for_each_entry(node,&leds,list){
		struct gpio_led_data *data;

		DBG(1, "value = [%s]",node->val);

		data = malloc(sizeof(struct gpio_led_data));
		memset(data,0,sizeof(struct gpio_led_data));

		data->led.name = node->val;

		s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "addr");
		DBG(1, "addr = [%s]", s);
		if (s) {
			data->addr =  strtol(s,0,0);
		}

		s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "delay");
		DBG(1, "delay = [%s]", s);
		if (s) {
			data->delay =  strtol(s,0,0);
		}

		s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "mode");
		DBG(1, "mode = [%s]", s);
		if (s) {

			if (!strncasecmp("direct",s,6))
				data->mode =  DIRECT;
			else if (!strncasecmp("linux",s,5))
				data->mode =  GPIO_LINUX;
			else if (!strncasecmp("csr",s,3))
				data->mode =  SHIFTREG_GPIO;
			else if (!strncasecmp("led_linux",s,9))
				data->mode =  LED_LINUX;
			else if (!strncasecmp("led_sk9822",s,10))
				data->mode =  LED_SK9822;
			else
				DBG(1, "Mode %s : Not supported!", s);
		}

		s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "active");
		DBG(1, "active = [%s]", s);
		if (s) {
			data->active = UNKNOWN;
			if (!strncasecmp("hi",s,1))
				data->active = HI;
			if (!strncasecmp("low",s,3))
				data->active = LOW;
		}

		data->led.func = &func;
		data->led.priv = data;

		switch (data->mode) {
			case LED_LINUX:
				s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "sysfs_attr");
				DBG(1, "sysfs_attr = [%s]", s);
				if (s) {
					data->sysfs_attr =  strdup(s);
				}
				data->brightness = 255;
				break;
			case LED_SK9822:
				s = ucix_get_option(s_ctx->uci_ctx, "hw" , data->led.name, "rgb_color");
				DBG(1, "rgb_color = [%s]", s);
				if (s) {
					if (strlen(s) == 6)
						data->rgb_color =  strdup(s);
					else
						DBG(1, "invalid rgb_color [%s]", s);
				}
				data->brightness = SK9822_DEFAULT_BRIGHTNESS;
				break;
			default:
				gpio_linux_output_init(data->addr);
				break;
		}

		led_add(&data->led);
	}

	if (gpio_shiftreg_clk != -1)
		gpio_shift_register_init(&led_gpio_shift_register, gpio_shiftreg_clk, gpio_shiftreg_dat, gpio_shiftreg_mask, gpio_shiftreg_bits);
}
